
#include "FastMapDistEucl.h"


/*===================================================*/
VECTOR *vecalloc( int n){
    VECTOR *pvec;

    assert( n > 0 );
    pvec = (VECTOR *) malloc( sizeof(VECTOR) );
    assert( pvec != NULL );
    pvec-> vec = (NUMBER *) malloc ( sizeof(NUMBER) * n );
    assert( (pvec->vec) != NULL );
    pvec->len = n;

    return(pvec);
}

void vecfree( VECTOR *pvec){
    free( (char *) pvec->vec);
    pvec->len =0;
    free( (char *) pvec);
}

NUMBER vecdist2( VECTOR *p1, VECTOR *p2){
    NUMBER res2; /* squared result */
    NUMBER v1, v2;
    int pos, length;

    assert( (p1->len) == (p2->len) );
    length = p1->len;
    res2=0.0;
    for(pos=0; pos< length; pos++){
	v1 = (p1->vec)[pos];
	v2 = (p2->vec)[pos];
	res2 += (v1-v2)*(v1-v2);
    }

    return( res2);
}

void vecput( VECTOR *pvec, int pos, NUMBER val){
    assert( pos >=0) ;
    assert( pos < (pvec->len)  );
    (pvec->vec)[pos] =  val;
}

NUMBER vecget( VECTOR *pvec, int pos){
    assert( pos >=0) ;
    assert( pos < (pvec->len)  );
    return( (pvec->vec)[pos] );

}

VECTOR *veccopy( VECTOR *psource){
    VECTOR *ptarget;
    int pos, length;

    assert( (psource->len) >0 );
    length = psource->len;
    ptarget = vecalloc(length);
    /* printf( "%d\n" , ptarget->len); */
    for(pos=0; pos<length; pos++){
	(ptarget->vec)[pos] = (psource->vec)[pos];
    }
    return( ptarget);
}

void vecprint( VECTOR *p){
    int i, length;

    /*  assert( (p->len) > 0 ); */
    length = p->len;

    for(i=0; i< length; i++){
	printf("%g\t", vecget(p, i));
    }
    printf("\n");
}
/*===================================================*/


/*===================================================*/
OBJECT *oballoc( int n){
    OBJECT *pob;

    assert( n > 0 );
    pob = (OBJECT *) malloc( sizeof(OBJECT) );
    assert( pob != NULL );
    pob-> vec = (NUMBER *) malloc ( n * sizeof(NUMBER) );
    assert( (pob->vec) != NULL );
    pob->len = n;

    return(pob);
}

void obfree( OBJECT *pob){
    free( (char *) pob->vec);
    pob->len =0;
    free( (char *) pob);
}

NUMBER obdist2( OBJECT *p1, OBJECT *p2){
     NUMBER res2; /* squared result */
     NUMBER v1, v2;
     int pos, length;

     assert( (p1->len) == (p2->len) );
     length = p1->len;
     res2=0.0;
     for(pos=0; pos< length; pos++){
 	v1 = (p1->vec)[pos];
 	v2 = (p2->vec)[pos];
   	res2 += (v1-v2)*(v1-v2);
     }
     return( res2);
 }

void obput( OBJECT *pob, int pos, NUMBER val){
    assert( pos >=0) ;
    assert( pos < (pob->len)  );
    (pob->vec)[pos] =  val;
}

NUMBER obget( OBJECT *pob, int pos){
    assert( pos >=0) ;
    assert( pos < (pob->len)  );
    return( (pob->vec)[pos] );

}

OBJECT *obcopy( OBJECT *psource){
    OBJECT *ptarget;
    int pos, length;

    assert( (psource->len) >0 );
    length = psource->len;
    ptarget = oballoc(length);
    for(pos=0; pos<length; pos++){
	(ptarget->vec)[pos] = (psource->vec)[pos];
    }
    return( ptarget);
}

void obprint( OBJECT *p){
    int i, length;

    /*  assert( (p->len) > 0 ); */
    length = p->len;

    for(i=0; i< length; i++){
	printf("%g\t", obget(p, i));
    }
    printf("\n");
}

OBJECT *line2ob(char *line){
    char *tmpline;  /* temp copy */
    OBJECT *pob;
    char *token;
    int ntokens;
    NUMBER val;
    int i;

    ntokens = counttokens(line);
    pob = oballoc(ntokens);

    /* copy, because strtok() messed up tmpline */
    tmpline=strdup(line);

    assert(ntokens>0); /* no empty lines */
    i=0;
    token = strtok(tmpline, BLANKS);
    /* atof did not work! */
    /* val = (NUMBER) atof( token); */
    sscanf( token, "%lf", &val);
    obput(pob, i, val);
    for(i=1; i< ntokens; i++){
        token = strtok( (char *)NULL, BLANKS);
    	sscanf( token, "%lf", &val);
	obput(pob, i,val);
    }
    free(tmpline);
    return(pob);
}
/*===================================================*/


/*===================================================*/
MATRIX *matrixalloc( int n){
    MATRIX *pmatrix;

    assert( n > 0 );
    pmatrix = (MATRIX *) malloc( sizeof(MATRIX) );
    assert( pmatrix != NULL );
    pmatrix-> list = (VECTOR **) malloc ( sizeof(VECTOR *) * n );
    assert( (pmatrix->list) != NULL );

    pmatrix->count = n;

    /* ******** DO NOT allocate space for vec ****
    for(i=0; i<count; i++){
        (pmatrix->list)[i] = vecalloc(m);
    }
    ******** */

    return(pmatrix);
}

void matrixfree( MATRIX *pmatrix){
    int i;
    for(i=0; i< (pmatrix->count); i++){
	vecfree( (pmatrix->list)[i] );
    }
    free( (char *) pmatrix->list);
    pmatrix->count =0;
    free( (char *) pmatrix);
}

void matrixput( MATRIX *pmatrix, int pos, VECTOR *pvec){
    assert( pos >=0) ;
    assert( pos < (pmatrix->count)  );
    (pmatrix->list)[pos] = veccopy(pvec);
}

void matrix2put( MATRIX *pmatrix, int row, int column, NUMBER val){
    VECTOR *pvec;
    assert( row >=0);
    assert( row < (pmatrix->count)  );
    pvec = (pmatrix->list)[row];
    vecput( pvec, column, val);
}

VECTOR *matrixget( MATRIX *pmatrix, int pos){
    assert( pos >=0) ;
    assert( pos < (pmatrix->count)  );
    return( (pmatrix->list)[pos] );
}

NUMBER matrix2get( MATRIX *pmatrix, int row, int column){
    VECTOR *pvec;
    NUMBER val;
    assert( row >=0);
    assert( row < (pmatrix->count)  );
    pvec = (pmatrix->list)[row];
    val = vecget( pvec, column);
    return(val);
}

void matrixprint( MATRIX *pmatrix){
    int i, n;
    if( pmatrix == NULL ){
	printf("empty matrix\n");
	return;
    }
    assert( pmatrix->count >0 );
    n = pmatrix->count;
    for(i=0; i<n; i++){
	vecprint( matrixget( pmatrix, i) );
    }
}

/* prints pairwise squared distances */
void printvdist2(MATRIX *pmat){
    int i, j, n;

    n = pmat->count;
    for(i=0; i<n; i++){
	for(j=0; j<n; j++){
	    printf("%g\t", vecdist2( matrixget(pmat, i),
				     matrixget(pmat, j) ) );
	}
	printf("\n");
    }
}
/*===================================================*/


/*===================================================*/
OA *oaalloc( int n){
    OA *poa;

    assert( n > 0 );
    poa = (OA *) malloc( sizeof(OA) );
    assert( poa != NULL );
    poa-> list = (OBJECT **) malloc ( sizeof(OBJECT *) * n );
    assert( (poa->list) != NULL );

    poa->count = n;

    /* ******** DO NOT allocate space for objects ****
    for(i=0; i<count; i++){
        (poa->list)[i] = oballoc(m);
    }
    ******** */

    return(poa);
}

void oafree( OA *poa){
    int i;
    for(i=0; i< (poa->count); i++){
	obfree( (poa->list)[i] );
    }
    free( (char *) poa->list);
    poa->count =0;
    free( (char *) poa);
}

/* copies an object to the given position */
void oaput( OA *poa, int pos, OBJECT *pob){
    assert( pos >=0) ;
    assert( pos < (poa->count)  );
    (poa->list)[pos] = obcopy(pob);
}

OBJECT *oaget( OA *poa, int pos){
    assert( pos >=0) ;
    assert( pos < (poa->count)  );
    return( (poa->list)[pos] );
}

void oaprint( OA *poa){
    int i, n;
    assert( poa->count > 0 );
    n = poa->count;
    for(i=0; i<n; i++){
	obprint( oaget( poa, i ) );
    }
}

/*===================================================*/


/*===================================================*/
void myrewind(FILE *fp){
    if( ferror(fp) ){ clearerr(fp);}
    rewind(fp);
    return;
}

int countlines(FILE *fp){
    int nlcount = 0;
    int ichar;

    myrewind(fp);
    while( (ichar=fgetc(fp)) != EOF ){
	if( ichar == '\n' ){ nlcount ++ ;}
    }
    myrewind(fp);

    return(nlcount);
}

/* measures the number of columns of the first line */
int countcolumns(FILE *fp){
    char *line;

    myrewind(fp);

    line= getline(fp);
    return( counttokens( line) );
}

void checklength(char *line){
    if( index(line, '\n') == NULL ){ /* too long */
	fprintf(stderr, "too long line - stopping \n");
	exit(-1);
    }
}

int counttokens(char *line){
    char *templine;
    int res = 0 ;

    /* create a copy, because strtok() messes its arg */
    templine = strdup(line);

    if( strtok(templine, BLANKS) != NULL){
	res ++ ;
    }
    while( strtok( (char *) NULL, BLANKS ) != NULL){
	res ++ ;
    }
    free( templine);
    return(res);
}

/* reads a line from fp, and checks its length */
char *getline(FILE *fp){
    static char line[MAXLINE];

    if( fgets(line, MAXLINE, fp) != '\0'){ /* EOF */
        checklength(line);
    }
    return(line);
}

/* checks to see if all the lines of the file have the same
   number of tokens */

BOOLEAN checkfile(FILE *fp){
    char *line;
    int nold, nnew;
    int linecount;
    BOOLEAN kosher = TRUE;

    myrewind(fp);

    /* read in the first line */
    line= getline(fp);
    linecount =1;
    if( feof(fp) ){
      return(-1); /*empty file */
    }else{
	nold = counttokens(line);
#ifdef DEBUG
	/* printf("%d:\t%s", nold, line); */
#endif
    }

    line = getline(fp);
    while( !feof(fp) ){
	linecount ++;
        nnew = counttokens(line);
#ifdef DEBUG
	/* printf("%d:\t%s", nnew, line); */
#endif
	if( nnew != nold){ 
	   kosher = FALSE;
	   fprintf(stderr, "line %d has %d tokens; line %d has %d\n",
	      1, nold, linecount, nnew);
	}
	/* assert( nnew == nold ); */
	line = getline(fp);
    }

    myrewind(fp);
    return (kosher);
}

/*===================================================*/









/*===================================================*/
/********************************************
   given an array of objects `*poa'
   and a desirable dimensionality `k'
   it returns an n*k matrix with coordinates
   of points, whose pairwise distances approximate
   the distances of the objects of `*poa'
 ***********************************************/

MATRIX *fastmap( OA *poa, int k){
   int n = poa->count;
   int i, a1, a2;
   int dims;
   NUMBER d12sq, d1xsq, d2xsq, x;
   MATRIX *pmat;
   VECTOR *pvec;

   pmat = matrixalloc(n);

   /* blank out *pmat */
   pvec = vecalloc( k );
   for( i=0; i<k; i++){
       /* blank out vector */
       vecput( pvec, i, 0.0);
   }
   for(i=0; i<n; i++){
       /* fill up matrix */
       matrixput( pmat, i, pvec);
   }
   vecfree(pvec);
   /* SHOULD MAKE THE ABOVE INTO  A FUNCTION */

   for( dims=0; dims<k; dims++){

       /* pick two remote objects */
       pickAnchors( poa, &a1, &a2, dims, pmat);

       /* compute projections of each object
	  on the axis defined by the two anchors */

       d12sq = mydist2( poa, a1, a2, dims, pmat);
       if( d12sq == 0.0 ) { return(pmat); }
	   /* we've exceeded the dimensionality of the data space */

       for( i=0; i<n; i++){
	   d1xsq = mydist2( poa, i, a1, dims, pmat);
	   d2xsq = mydist2( poa, i, a2, dims, pmat);
	   x = (d1xsq + d12sq - d2xsq) / ( 2.0 * sqrt( (double) d12sq ) );
	   matrix2put( pmat, i, dims, x);
       } /* end for i */
   } /* end for dims */

   return( pmat );
}


/* choose to far-away objects */
void pickAnchors( OA * poa, int *pa1, int *pa2, int dims, MATRIX *pmat){
    int i,j;

    i=0;
    j = Furthest( poa, i, dims, pmat);
    i = Furthest( poa, j, dims, pmat);

    *pa1 = i;
    *pa2 = j;
    return;
}

/* for a given object 'i', returns the index of the furthest one */
int Furthest( OA *poa, int icurrent, int dims, MATRIX *pmat){
    NUMBER max = 0.0;
    int furthest= icurrent;
    int n = poa->count;
    int j;

    for(j=0; j<n; j++){
	if (mydist2( poa, icurrent, j, dims, pmat) > max){
	   max = mydist2( poa, icurrent, j, dims, pmat);
	   furthest = j;
        }
    }
    return(furthest);
}

///* squared distance of two objects, after dims-1 projections */
NUMBER mydist2( OA *poa, int o1, int o2, int dims, MATRIX *pmat){
    OBJECT *p1 = oaget( poa, o1);
    OBJECT *p2 = oaget( poa, o2);
    NUMBER res2 = obdist2( p1, p2);

    /* decrease it, by the projections */
    int i;
    for(i=0; i< dims; i++){
	NUMBER temp = matrix2get(pmat, o1, i) - matrix2get(pmat, o2, i);
	res2 -=  temp*temp;
    }
    return(res2);
}

/* prints the  squared distances of all pairs of elements */
void printdist2(OA *poa){
    int i, j, n;
    n = poa->count;
    for(i=0; i< n; i++){
	for(j=0; j<n; j++){
	    printf("%g\t", obdist2( oaget(poa, i), oaget(poa, j) ));
	}
	printf("\n");
    }
}

/* reads in a file, into an object array */
OA *file2oa(FILE *fp){
    OA *poa;
    OBJECT *pob;
    int i, n;
    char *line;

    myrewind(fp); 
    n = countlines(fp);

    assert( n>0); /* no empty file */

    poa = oaalloc(n);
    assert( poa != NULL);

    myrewind(fp);
    i=0;
    line = getline(fp);
    while( !feof(fp) ){
	pob = line2ob(line);
	oaput( poa, i, pob);
	obfree(pob);

	i++;
	line = getline(fp);
    }

    assert( i == n);
    for(i=0; i<n; i++){
	/*check the lengths of the objects for equality */
    }

    return(poa);
}
/*===================================================*/



/*===================================================*/
/*   MODIFICATION::: M. Mignotte::: 18 August 2010   */
/*===================================================*/

//Pour Distance de Bhattacharyaa

// squared distance of two objects, after dims-1 projections 
//NUMBER mydist2( OA *poa, int o1, int o2, int dims, MATRIX *pmat){
//    OBJECT *p1 = oaget( poa, o1);
//    OBJECT *p2 = oaget( poa, o2);
//    NUMBER res2 = obdist2( p1, p2);

//    /* decrease it, by the projections */
//    int i;
//    for(i=0; i< dims; i++){
//	NUMBER temp = matrix2get(pmat, o1, i) - matrix2get(pmat, o2, i);
//	res2 -=  temp*temp;
//
//        //Modification:Heuristic  Wang et al.
//        if (res2<0) res2 *= -1;
//    }
//    return( res2);
//}

//NUMBER obdist2( OBJECT *p1, OBJECT *p2){
//    float res2; 
//    float v1,v2;
//    float  Sum1,Sum2;
//    int pos,length;
//
//    assert( (p1->len) == (p2->len) );
//    length = p1->len;
//    
//    Sum1 = 0.0;
//    Sum2 = 0.0; 
//    for(pos=0; pos<length; pos++){
//	v1 = (p1->vec)[pos];
//	v2 = (p2->vec)[pos];
//      Sum1 += v1;
//      Sum2 += v2; }
//
//    res2 = 0.0;
//    for(pos=0; pos<length; pos++){
//	v1 = (p1->vec)[pos];
//	v2 = (p2->vec)[pos];
//	res2 += sqrt((v1/Sum1)*(v2/Sum2)); }
//    
//    if (res2>1.0) res2 = 1.0;
//    res2 = 10000*(1-res2);
//
//    return(res2);
//}

/* reads in a matrix[0...n][0...Dim], into an object array */
OA *Mat2oa(float** Mat,int n,int dim){
    OA *poa;
    OBJECT *pob;
    int k;

    poa=oaalloc(n);
    assert(poa != NULL);

    for(k=0;k<n;k++) 
      { pob=Mat2ob(Mat,k,dim);
	oaput(poa,k,pob);
	obfree(pob); }

    return(poa);
}

/* reads each sample, into an object */
OBJECT *Mat2ob(float** Mat,int numech,int dim){
    OBJECT *pob;
    int k;

    pob = oballoc(dim);
    for(k=0;k<dim;k++) obput(pob,k,Mat[numech][k]);
    return(pob);
}

/* convert a MATRIX into mat[0...n][0...dim] */
void convertmat(MATRIX *pmatrix,float** Mat){
    int k,n;
   
    assert(pmatrix->count>0);
    n=pmatrix->count;
    for(k=0;k<n;k++) fillMat(matrixget(pmatrix,k),Mat,k);   
}


void fillMat(VECTOR *p,float** Mat,int n){
    int k,length;

    length=p->len;
    for(k=0;k<length;k++) Mat[n][k]=vecget(p,k);
}

/* FastMap  Matin[0...n][0...dim_in] >  Matout[0...n][0...dim_out] */
void MyFastMap(float** Matin,float** Matout,int lgth,int dimin,int dimout,int verb) {
   MATRIX *mat;
   OA *a;

   a=Mat2oa(Matin,lgth,dimin);
   mat=fastmap(a,dimout);
   
   if (verb) matrixprint(mat);
   convertmat(mat,Matout);
   
   matrixfree(mat);
   oafree(a); 
}

/* FastMap  Matin[0...n][0...dim_in] >  Matin[0...n][0...dim_out] */
void MyFastMap_(float** Matin,int lgth,int dimin,int dimout,int verb) {
   MATRIX *mat;
   OA *a;

   a=Mat2oa(Matin,lgth,dimin);
   mat=fastmap(a,dimout);
   
   if (verb) matrixprint(mat);
   convertmat(mat,Matin);
   
   matrixfree(mat);
   oafree(a); 
}

/*===================================================*/

