//---------------------------------------------------
// module  : f_DisplayHyper.cc                       
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "bcocdm.h"

//------------------------------------------------
// METHODES DE LA CLASSE DISPLAYHYPER  -------------
//------------------------------------------------
//----------------------------------------------------------
// Constructeur de la classe DisplayHyper
//----------------------------------------------------------
DisplayHyper::DisplayHyper()           
 {
  //Initialisation
  depth=DEPTH;
  length=LENGTH;
  width=WIDTH;

  //Allocation Memoire Matrice  
  ImgCube=simatrix_allocate_3d(depth,length,width);
  ImgCoul=imatrix_allocate_3d(TROIS,length,width);

  //>Info
  BandInfo=fmatrix_allocate_2d(depth,4);

  //>Cube
  CubLabRgbLab=simatrix_allocate_4d(LMAX-LMIN,AMAX-AMIN,BMAX-BMIN,3);

  //>Gradient
  MatxGrd=fmatrix_allocate_4d(10,TROIS,length,width);

  //Parametre
  Correlation=0.0;

  //Flag
  flag_corr=0;
 }


//----------------------------------------------------------
// Destructeur de la classe DisplayHyper
//----------------------------------------------------------
DisplayHyper::~DisplayHyper()
{ 
  //Libere Memoire Matrice 
  if (ImgCube)    free_simatrix_3d(ImgCube,depth); 
  if (ImgCoul)    free_imatrix_3d(ImgCoul,TROIS);
  if (BandInfo)   free_fmatrix_2d(BandInfo);
}

//------------//
//--METHODES--//
//------------//
//----------------------------------------------------------
// FasMap + METROPOLIS
//
//----------------------------------------------------------    
void DisplayHyper::Display()
{
 int i,j,k,l;
 int DeltaBand;
 int B1,B2,B3,B4,B5,B6;
 float Nrj,NrjMin,NrjPrec;
 float k_0,k_1,k_2;
 float IndexF;

 //>Decoupage Bande
 DeltaBand=(int)((float)depth/3.0);
 B1=0;
 B2=B3=DeltaBand;
 B4=B5=B3+DeltaBand;
 B6=depth-1; 

 //>TimeObjet
 Chrono  chrono; 
 chrono.reset();

 //>Information
 printf("\n ALGORITHM ...BCOCDM");
 printf("\n -------------------");
 printf("\n Image  : ..... [%s]",Name_Img);
 printf(" [Ext:%s]",Extension);
 printf("  Size:[depth:%d | lgth:%d | wdth:%d]",depth,length,width);
 printf("\n Bandes :...... [%d-%d] [%d-%d] [%d-%d]",B1,B2,B3,B4,B5,B6);
 printf("\n ---------");
 printf("\n Regul:[%.3f]",Regul);
 printf("\n ---------");
 printf("\n");
 fflush(stdout);

 //>AllocationMemoire
 float*** ImgLAB=fmatrix_allocate_3d(TROIS,length,width);
 float*** ImgRVB=fmatrix_allocate_3d(TROIS,length,width);

 float**  Matx=fmatrix_allocate_2d((length*width),depth);
 float*** ImgTmp=fmatrix_allocate_3d(TROIS,length,width);
 float*** ImgRes=fmatrix_allocate_3d(TROIS,length,width);
 short int*** NonLocGrph=simatrix_allocate_3d(TROIS,NBCNX,(width*length));

 //====
 // 0 =
 //====
 //===========================================
 //   ==========================
 //===========================================
 if (1)
    {
     printf("\n Precompute...");
     fflush(stdout);
     FillCubeLabRgbLab();
     TimeSec=chrono.elapsed(SECOND); chrono.print(); 
     fflush(stdout);
    }
 

 //===========================================
 // INIT ==============================================
 //===========================================
 //>Chargement & Visu
 printf("\n Chargement AVIRIS Img...");
 LoadImgHyperspectral(Name_Img,ImgCube,depth,length,width);
 SelectInfoBands(ImgCube,depth,length,width,0,BandInfo,RSBMIN);  

 //====
 // 1 =
 //====
 //===========================================
 // FastMap Iinitialization  ==========================
 //===========================================-
  if (0)
   {
    k=0;
    for(i=0;i<length;i++) for(j=0;j<width;j++) { for(l=0;l<depth;l++) Matx[k][l]=ImgCube[l][i][j]; k++; }
    printf("\n FasMap[LAB]");
    MyFastMap_(Matx,(length*width),depth,3,0);
    for(k=0;k<(length*width);k++)
         {  i=(int)(k/width); j=(int)(k%width);
            ImgLAB[0][i][j]=Matx[k][0];
            ImgLAB[1][i][j]=Matx[k][1];
            ImgLAB[2][i][j]=Matx[k][2]; }
   }


 if (1)
    {
     //>L Matrix
     k=0;
     for(i=0;i<length;i++) for(j=0;j<width;j++) { for(l=B1;l<B2;l++) Matx[k][l]=ImgCube[l][i][j]; k++; }
     printf("\n FasMap[L]");
     MyFastMap_(Matx,(length*width),B2-B1,1,0);
     for(k=0;k<(length*width);k++) {  i=(int)(k/width); j=(int)(k%width); ImgLAB[0][i][j]=Matx[k][0]; }

     //>A Matrix
     k=0;
     for(i=0;i<length;i++) for(j=0;j<width;j++) { for(l=B3;l<B4;l++) Matx[k][l-B3]=ImgCube[l][i][j]; k++; }
     printf("\n FasMap[A]");
     MyFastMap_(Matx,(length*width),B4-B3,1,0);
     for(k=0;k<(length*width);k++) {  i=(int)(k/width); j=(int)(k%width); ImgLAB[1][i][j]=Matx[k][0]; }

     //>B Matrix
     k=0;
     for(i=0;i<length;i++) for(j=0;j<width;j++) { for(l=B5;l<B6;l++) Matx[k][l-B5]=ImgCube[l][i][j]; k++; }
     printf("\n FasMap[B]\n");
     MyFastMap_(Matx,(length*width),B6-B5,1,0);
     for(k=0;k<(length*width);k++) {  i=(int)(k/width); j=(int)(k%width); ImgLAB[2][i][j]=Matx[k][0]; } 
    }

 //====
 // 2 =
 //====
 //===============================================
 // ConversionFinale LAB->RGB & Recalage LAB ==========================
 //===============================================
     CopyMat(ImgTmp,ImgLAB,length,width,TROIS);
     GoodRangeLABvalues(ImgLAB,length,width,1.0);
     ComputeSatPix(ImgLAB,length,width); 
     k_0=1.0/ComputeNrjPairwise(ImgTmp,ImgLAB,length,width,10);
     printf("\n [k_0=%.3f]",k_0);
  
 //====
 // 3 =
 //====
 //===============================================
 // Global estimation  -k- =========================================
 //===============================================
 if (1)
    {
    //>NonLocalMultiResGraph
    printf("\n  Local Graph ...");
    ComputeNonLocalGraph(ImgCube,B1,B6,length,width,NonLocGrph,SZW,NBCNX);
 
    //>Estimation -k-
    IndexF=-1;
    k_1=-1;
    Nrj=BIGNUMBER;
    NrjMin=BIGNUMBER;
    for(k=0;k<100;k++)
       { IndexF=1.0+(float)(k)/10.0;
         CopyMat(ImgTmp,ImgLAB,length,width,TROIS);
         GoodRangeLABvalues(ImgTmp,length,width,IndexF);
         NrjPrec=Nrj;
         Nrj=ComputeNrjGraph(ImgTmp,length,width,NonLocGrph,NBCNX,Regul,k_0*IndexF);   
         if (!k) NrjMin=Nrj;
         if (Nrj<=NrjMin) { NrjMin=Nrj; k_1=IndexF; }     
         printf("  [%d][%.2f][Nrj=%.2f]",k,IndexF,Nrj);
         fflush(stdout); 
         if (Nrj>NrjPrec) break; }

    CopyMat(ImgTmp,ImgLAB,length,width,TROIS);
    GoodRangeLABvalues(ImgTmp,length,width,k_1);
    printf("\n   [k_0=%.3f][k_1=%.2f]",k_0,k_1);
    
    //>Least Square Correction
    k_2=ComputeScalingFactor(ImgTmp,ImgCube,length,width,depth,k_0*k_1,20);
    printf("[k_2=%.5f]",k_2);
    }
   
 //====
 // 4 =
 //====
 //===============================================
 // Metropolis <Local Exploration>  
 //===============================================
 //
 //    float*** Img,float*** ImgRes,int lgth,int wdth,
 //    short int*** NonLocGrph,int NbCnx,int NbIter,
 //    float Temp_Init,float Radius,float Regul,float fact
 //
 // >>> ImgRes
 //
 //================================================
 if (1) Metropolis(ImgTmp,ImgLAB,length,width,NonLocGrph,NBCNX,Regul,k_0*k_1*k_2);
 TimeSec=chrono.elapsed(SECOND); chrono.print(); 
 fflush(stdout);

 //Conversion
 //==========
 Convert_LAB_TO_RGB(ImgLAB,length,width,ImgRVB); 

 //Validation Cor/Featuresep
 //=========================
 if (flag_corr)
    { Correlation=SAM_Valid(ImgRVB,ImgCube,depth,length,width,&FeatSep); 
      printf("\n  >> Correlation=%.4f",Correlation); }

 //FillInfo 
 //========
 FillInfoProcess();
 printf("  [%s]",InfoProcess);
 
 //Sauvegarde
 //==========
 if (flag_save)
    { printf("\n  Sauvegarde");
      fflush(stdout);
      Recal(ImgLAB,length,width);
      //SaveImagePpm("HypImg_LAB",Extension,InfoProcess,ImgLAB,length,width);
      SaveImagePpm("HypImg_RBV",Extension,InfoProcess,ImgRVB,length,width); }

  for(k=0;k<TROIS;k++) for(i=0;i<length;i++) for(j=0;j<width;j++) ImgCoul[k][i][j]=ImgRVB[k][i][j];


  //>Liberation Memoire
  //===================
  if (ImgLAB)   free_fmatrix_3d(ImgLAB,TROIS); 
  if (ImgRVB)   free_fmatrix_3d(ImgRVB,TROIS); 
  if (Matx)     free_fmatrix_2d(Matx);
  if (ImgTmp)   free_fmatrix_3d(ImgTmp,TROIS); 
  if (ImgRes)   free_fmatrix_3d(ImgRes,TROIS);
}


