//---------------------------------------------------
// module  : f_DisplayHyper2.cc                       
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "bcocdm.h"


//----------------------------------------------------------
// Info 
//----------------------------------------------------------    
void DisplayHyper::FillInfoProcess()
{
 char Buff[NCHAR];


 //>Correlation
 sprintf(Buff,"_Regul%.0f",Regul*100);
 strcpy(InfoProcess,Buff);
 
 //>Correlation
 sprintf(Buff,"_Cor%.3f",Correlation);
 strcat(InfoProcess,Buff);

 //>FeatureSeparability
 sprintf(Buff,"_Con%.1f",FeatSep);
 strcat(InfoProcess,Buff);
}

//----------------------------------------------------------
//METHODES :: METROPOLIS -----------------------------------
//----------------------------------------------------------
//----------------------------------------------------------
// retourne un nombre aleatoire entre zero et un
//----------------------------------------------------------
float DisplayHyper::Randomize(void)
{ return ((float)rand()/RAND_MAX); }

//----------------------------------------------------------
// calcule le nombre de differences a chaque iteration
//----------------------------------------------------------
float DisplayHyper::DiffMat(float** pmatl1,float** pmatl2,int lgth,int wdth,float precs)
{
 int i,j;
 float chg;

 //>Compte NbChgt
 chg=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) if (fabs(pmatl1[i][j]-pmatl2[i][j])>precs) chg++; 
 chg/=(lgth*wdth);
 chg*=100.0;
       
 return chg;
}

//----------------------------------------------------------
//----------------------------------------------------------

//----------------------------------------------------------
// Compute Global NRJ
//----------------------------------------------------------
//
//----------------------------------------------------------     
float DisplayHyper::ComputeNrj(float*** Img,int lgth,int wdth,short int*** NonLocGrph,int NbCnx,float Regul,float fact)
{
 int i,j,k;
 float tmp0,tmp1,tmp2;
 float Nrj;
 int posr,posc;
 int nbp;
 float dist;
 float us_l,us_a,us_b;
 float ut_l,ut_a,ut_b;
 int   usl,usa,usb;
 int   utl,uta,utb;

 //Variation Totale
 Nrj=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
     {
      us_l=Img[0][i][j];
      us_a=Img[1][i][j];
      us_b=Img[2][i][j];
      usl=(int)(us_l/FACT_CUBE);
      usa=(int)(us_a/FACT_CUBE);
      usb=(int)(us_b/FACT_CUBE);
      if (usl>=LMAX) usl=LMAX-1;
      if (usl<LMIN)  usl=LMIN;
      if (usa>=AMAX) usa=AMAX-1;
      if (usa<AMIN)  usa=AMIN;
      if (usb>=BMAX) usb=BMAX-1;
      if (usb<BMIN)  usb=BMIN;
      us_l=CubLabRgbLab[usl-LMIN][usa-AMIN][usb-BMIN][0];
      us_a=CubLabRgbLab[usl-LMIN][usa-AMIN][usb-BMIN][1];
      us_b=CubLabRgbLab[usl-LMIN][usa-AMIN][usb-BMIN][2];

      tmp0=0.0;
      tmp1=0.0;
      tmp2=0.0;
      for(k=0;k<NbCnx;k++)
        { nbp=(i*wdth)+j;
          posr=(int)NonLocGrph[0][k][nbp];
          posc=(int)NonLocGrph[1][k][nbp];
          dist=NonLocGrph[2][k][nbp];

          ut_l=Img[0][posr][posc];
          ut_a=Img[1][posr][posc];
          ut_b=Img[2][posr][posc];
          utl=(int)(ut_l/FACT_CUBE);
          uta=(int)(ut_a/FACT_CUBE);
          utb=(int)(ut_b/FACT_CUBE);
          if (utl>=LMAX) utl=LMAX-1;
          if (utl<LMIN)  utl=LMIN;
          if (uta>=AMAX) uta=AMAX-1;
          if (uta<AMIN)  uta=AMIN;
          if (utb>=BMAX) utb=BMAX-1;
          if (utb<BMIN)  utb=BMIN;
          ut_l=CubLabRgbLab[utl-LMIN][uta-AMIN][utb-BMIN][0];
          ut_a=CubLabRgbLab[utl-LMIN][uta-AMIN][utb-BMIN][1];
          ut_b=CubLabRgbLab[utl-LMIN][uta-AMIN][utb-BMIN][2];
    
          tmp0=CARRE(us_l-ut_l)+CARRE(us_a-ut_a)+CARRE(us_b-ut_b);
          tmp1=CARRE(fact*dist-sqrt(tmp0));
          tmp2+=tmp1-(Regul*tmp0); }

      Nrj+=tmp2/(float)NbCnx;
     }

 //Retour
 return Nrj/(float)(wdth*lgth);
}

//----------------------------------------------------------
// Compute Local NRJ  
//----------------------------------------------------------
//
//---------------------------------------------------------- 
float DisplayHyper::ComputeLocNrj(float*** Img,int row,int col,float* val,\
int wdth,short int*** NonLocGrph,int NbCnx,float Regul,float fact)
{
 int k;
 float tmp0,tmp1,tmp2;
 float Nrj;
 int posr,posc;
 int nbp;
 float dist;
 float us_l,us_a,us_b;
 float ut_l,ut_a,ut_b;
 int   usl,usa,usb;
 int   utl,uta,utb;

 //Energie Locale
 us_l=val[0];
 us_a=val[1];
 us_b=val[2]; 
 usl=(int)(us_l/FACT_CUBE);
 usa=(int)(us_a/FACT_CUBE);
 usb=(int)(us_b/FACT_CUBE);
 if (usl>=LMAX) usl=LMAX-1;
 if (usl<LMIN)  usl=LMIN;
 if (usa>=AMAX) usa=AMAX-1;
 if (usa<AMIN)  usa=AMIN;
 if (usb>=BMAX) usb=BMAX-1;
 if (usb<BMIN)  usb=BMIN;
 us_l=CubLabRgbLab[usl-LMIN][usa-AMIN][usb-BMIN][0];
 us_a=CubLabRgbLab[usl-LMIN][usa-AMIN][usb-BMIN][1];
 us_b=CubLabRgbLab[usl-LMIN][usa-AMIN][usb-BMIN][2];

 tmp0=tmp1=tmp2=0.0;
 for(k=0;k<NbCnx;k++)
     { nbp=(row*wdth)+col;
       posr=(int)NonLocGrph[0][k][nbp];
       posc=(int)NonLocGrph[1][k][nbp];
       dist=NonLocGrph[2][k][nbp];

       ut_l=Img[0][posr][posc];
       ut_a=Img[1][posr][posc];
       ut_b=Img[2][posr][posc];        
       utl=(int)(ut_l/FACT_CUBE);
       uta=(int)(ut_a/FACT_CUBE);
       utb=(int)(ut_b/FACT_CUBE);
       if (utl>=LMAX) utl=LMAX-1;
       if (utl<LMIN)  utl=LMIN;
       if (uta>=AMAX) uta=AMAX-1;
       if (uta<AMIN)  uta=AMIN;
       if (utb>=BMAX) utb=BMAX-1;
       if (utb<BMIN)  utb=BMIN;
       ut_l=CubLabRgbLab[utl-LMIN][uta-AMIN][utb-BMIN][0];
       ut_a=CubLabRgbLab[utl-LMIN][uta-AMIN][utb-BMIN][1];
       ut_b=CubLabRgbLab[utl-LMIN][uta-AMIN][utb-BMIN][2];
    
       tmp0=CARRE(us_l-ut_l)+CARRE(us_a-ut_a)+CARRE(us_b-ut_b);
       tmp1=CARRE(fact*dist-sqrt(tmp0));
       tmp2+=tmp1-(Regul*tmp0); }

  Nrj=tmp2/(float)NbCnx; 

 //Retour
 return Nrj;
}

//----------------------------------------------------------
//----------------------------------------------------------
// Metropolis
//----------------------------------------------------------
//
//----------------------------------------------------------    
void DisplayHyper::Metropolis(float*** Img,float*** ImgRes,int lgth,int wdth,\
short int*** NonLocGrph,int NbCnx,float Regul,float fact)
{
 int i,j,k,l,m;
 float nrj,nrj1,nrj2;
 float ValImg;
 float rd,rd2; 
 float DeltaNrj;
 float ProbAcc;
 float RadiusDiv2; 
 float nbchgt;
 float Temp,TempInv;
 int SzW;

 //>Parameter <METROPOLIS>
 float Temp_Fina=TEMPMET_FINAL;
 float Temp_Init=TEMPMET_INIT;
 float Radius=RADIUS_MET;
 int   Kmax=NBIT_METR;
  
 //AllocationMemory
 float* TabVal1=fmatrix_allocate_1d(TROIS);
 float* TabVal2=fmatrix_allocate_1d(TROIS);

 //>Init
 RadiusDiv2=Radius/2.0;
 SzW=(int)sqrt(NbCnx);
 for(l=0;l<TROIS;l++) CopyMat(MatxGrd[1][l],Img[l],lgth,wdth);
 for(l=0;l<TROIS;l++) CopyMat(MatxGrd[2][l],Img[l],lgth,wdth);
 nrj=ComputeNrj(MatxGrd[1],lgth,wdth,NonLocGrph,NbCnx,Regul,fact);
 printf("\n    > METROPOLIS::[NbIter=%d][Rad=%.1f]",Kmax,Radius);  
 printf("[TpInit=%.1f][TpFinal=%.4f]",Temp_Init,Temp_Fina);
 printf("\n                  [Regul=%.3f][fact=%.3f]",Regul,fact);
 printf("\n\n     >  <NrjInit=%.2f>",nrj);

  //====================================================================
  //METROPOLIS >>> sur Img 
  //====================================================================
  for(k=0;;k++)
     {
      Temp=Temp_Init*powf((Temp_Fina/Temp_Init),((float)k/Kmax));
      TempInv=1.0/Temp;
      if (Temp<Temp_Fina) break;

      if (!(k%10)) 
      for(l=0;l<TROIS;l++) CopyMat(MatxGrd[2][l],MatxGrd[1][l],lgth,wdth);

      //-----------------------------------------------------------
      // Loop Img > Loop Pixels
      //------------------------------------------------------------
          for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) for(l=0;l<TROIS;l++)
            {
               //Gibbs Sampling Record
	       ValImg=MatxGrd[1][l][i][j];
  	       rd=ValImg-RadiusDiv2+(Radius*Randomize());
 
               if (rd<0.0)   rd=0.0;
               if (rd>255.0) rd=255.0;

	        for(m=0;m<TROIS;m++) { TabVal1[m]=MatxGrd[1][m][i][j]; TabVal2[m]=MatxGrd[1][m][i][j]; }
                TabVal2[l]=rd;
 
                nrj1=ComputeLocNrj(MatxGrd[1],i,j,TabVal1,wdth,NonLocGrph,NbCnx,Regul,fact);
                nrj2=ComputeLocNrj(MatxGrd[1],i,j,TabVal2,wdth,NonLocGrph,NbCnx,Regul,fact);               
                DeltaNrj=nrj2-nrj1;

                if (DeltaNrj<0) MatxGrd[1][l][i][j]=rd;

                else { ProbAcc=exp(-TempInv*DeltaNrj); 
                       rd2=Randomize();
                       if (rd2<ProbAcc)  MatxGrd[1][l][i][j]=rd; }    
           
             }//End_Loop_Pixel------------------------------------------  

          //>Info
          if (!(k%10))
	    {  nrj=ComputeNrj(MatxGrd[1],lgth,wdth,NonLocGrph,NbCnx,Regul,fact); 
               nbchgt=DiffMat(MatxGrd[2][0],MatxGrd[1][0],lgth,wdth,PREC_MET);
               printf("\n  [%d]>%.2f [Temp:%.4f][NbChgt/100=%.2f]",k,nrj,Temp,nbchgt); }

         }//End_Simulated_Annealing==================================================================
  
 //Nrj_finale[Img1] 
 for(l=0;l<TROIS;l++) CopyMat(ImgRes[l],MatxGrd[1][l],lgth,wdth);
 nrj=ComputeNrj(ImgRes,lgth,wdth,NonLocGrph,NbCnx,Regul,fact);
 NrjImg=nrj;
 printf("\n   >> [Nrj:%.3f]",nrj); 
 fflush(stdout);

 //FreeMemory
 if (TabVal1) free_fmatrix_1d(TabVal1);
 if (TabVal2) free_fmatrix_1d(TabVal2);
}

//=============================
//=============================
//    PRE-COMPUTE
//=============================
//=============================
//-----------------------------------------------
// FillCubeLabRgbLab LxAxB
// 
// [L][A][B][0]=l  [L][A][B][1]=a  [L][A][B][2]=b                        
//  
//  L=[LMIN:LMAX]   
//  A=[AMIN:AMAX]   
//  B=[BMIN:BMAX]
//
//-----------------------------------------------
void DisplayHyper::FillCubeLabRgbLab()
{
 int i,j,k;
 float r,g,bl;
 float l,a,b;

 //Boucle
 for(k=LMIN;k<LMAX;k++) for(i=AMIN;i<AMAX;i++) for(j=BMIN;j<BMAX;j++)
   {    
    //>Conversion
    LABToRGB(k*FACT_CUBE,i*FACT_CUBE,j*FACT_CUBE,&r,&g,&bl);
    
    //>Saturation
    if (r<0)     { r=0;      }
    if (g<0)     { g=0.0;    }
    if (bl<0)    { bl=0.0;   }
    if (r>255)   { r=255.0;  }
    if (g>255)   { g=255.0;  }
    if (bl>255)  { bl=255.0; }

    //Conversion inverse
    RGBToLAB(r,g,bl,&l,&a,&b);
 
    //Record
    CubLabRgbLab[k-LMIN][i-AMIN][j-BMIN][0]=(short int)l;
    CubLabRgbLab[k-LMIN][i-AMIN][j-BMIN][1]=(short int)a;
    CubLabRgbLab[k-LMIN][i-AMIN][j-BMIN][2]=(short int)b;
   }
}

//=============================
//=============================
//    STEP -3-
//=============================
//=============================
//-----------------------------------------------
// ComputeNrjPairwise
//-----------------------------------------------
float DisplayHyper::ComputeNrjPairwise(float*** ImgIn,float *** ImgOut,int lgth,int wdth,int sub)
{
  int i,j,k,l;
  float ValLocR,ValLocV,ValLocB;
  float ValLocR_,ValLocV_,ValLocB_;
  float NrjPairIn,NrjPairOut;
  float tmp;
  float fact;

  NrjPairIn=0.0;
  NrjPairOut=0.0;
  for(i=0;i<lgth;i+=sub) for(j=0;j<wdth;j+=sub)
    { ValLocR=ImgIn[0][i][j];
      ValLocV=ImgIn[1][i][j];
      ValLocB=ImgIn[2][i][j];

      ValLocR_=ImgOut[0][i][j];
      ValLocV_=ImgOut[1][i][j];
      ValLocB_=ImgOut[2][i][j];

      tmp=0.0;
      for(k=0;k<lgth;k+=sub) for(l=0;l<wdth;l+=sub)
        tmp+=CARRE(ValLocR-ImgIn[0][k][l])+CARRE(ValLocV-ImgIn[1][k][l])+CARRE(ValLocB-ImgIn[2][k][l]);

      NrjPairIn+=sqrt(tmp);
 
      tmp=0.0;
      for(k=0;k<lgth;k+=sub) for(l=0;l<wdth;l+=sub)
        tmp+=CARRE(ValLocR_-ImgOut[0][k][l])+CARRE(ValLocV_-ImgOut[1][k][l])+CARRE(ValLocB_-ImgOut[2][k][l]);

      NrjPairOut+=sqrt(tmp); }

  fact=NrjPairIn/NrjPairOut;  
  return fact;
}

//----------------------------------------------------------
// ComputeScalingFactor                           
//----------------------------------------------------------
float DisplayHyper::ComputeScalingFactor(float*** Img,short int*** HypImg,int lgth,int wdth,int NbFeat,float fact,int Sub)
{
  int i,j,k,l,m;
  float EstDist,TrueDist;
  float ScaleNum,ScaleDen;
  float ScaleFact;

  //Constante
  const float PREC=0.000001;
 
  //>SCALE_Factor
  //=========================================
  ScaleNum=ScaleDen=0.0;
  for(i=0;i<lgth;i+=Sub) for(j=0;j<wdth;j+=Sub) for(k=0;k<lgth;k+=Sub) for(l=0;l<wdth;l+=Sub)
       { 
         EstDist=0.0;   for(m=0;m<3;m++)      EstDist+=CARRE(Img[m][i][j]-Img[m][k][l]);
         TrueDist=0.0;  for(m=0;m<NbFeat;m++) TrueDist+=CARRE(HypImg[m][i][j]-HypImg[m][k][l]);
       
         EstDist+=PREC;
         TrueDist+=PREC;

         ScaleNum+=(sqrt(EstDist)/(fact*sqrt(TrueDist))); 
         ScaleDen+=(EstDist/(CARRE(fact)*TrueDist)); 
       }//=End=
     
  ScaleFact=(ScaleDen/ScaleNum);
   
  //>Return
  //=========================================
  return ScaleFact;
}

//-----------------------------------------------
// GoodRangeLABvalues
//  L in [0..100]
//  A,B -Amplitude 100 centre sur 0
//-----------------------------------------------
void DisplayHyper::GoodRangeLABvalues(float*** Img,int lgth,int wdth,float fact)
{
  int i,j;
  float moy;

 //Presentation
 printf("\n > [fact:%.2f]",fact);

 //Amplitude 100
 RecalI(Img[0],lgth,wdth,0.0,100.0*fact);
 RecalI(Img[1],lgth,wdth,0.0,100.0*fact);
 RecalI(Img[2],lgth,wdth,0.0,100.0*fact);

 moy=ComputeMean(Img[1],lgth,wdth);
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)  Img[1][i][j]-=moy;
 //Stat(Img[1],lgth,wdth);

 moy=ComputeMean(Img[2],lgth,wdth);
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)  Img[2][i][j]-=moy;
 //Stat(Img[2],lgth,wdth);  
}

//-----------------------------------------------
// ValLabRgbLab
//-----------------------------------------------
void DisplayHyper::ValLabRgbLab(float* l,float* a,float* b)
{
 float r,g,bl;

 //>Conversion
 LABToRGB((*l),(*a),(*b),&r,&g,&bl);
    
 //>Saturation
 if (r<0)     r=0.0;
 if (g<0)     g=0.0;
 if (bl<0)    bl=0.0;
 if (r>255)   r=255.0;
 if (g>255)   g=255.0;
 if (bl>255)  bl=255.0;

 //Conversion inverse
 RGBToLAB(r,g,bl,l,a,b);
}

//----------------------------------------------------------
// ComputeNrjGraph
// ===============
//
// float*** Img :: Image[lgth][wdth] LAB
//
//---------------------------------------------------------- 
float DisplayHyper::ComputeNrjGraph(float*** Img,int lgth,int wdth,short int*** NonLocGrph,int NbCnx,float Regul,float fact)
{
 int i,j,k;
 float tmp0,tmp1,tmp2;
 float Nrj;
 int posr,posc;
 int nbp;
 float dist;
 float us_l,us_a,us_b;
 float ut_l,ut_a,ut_b;
 int   usl,usa,usb;
 int   utl,uta,utb;

 //Variation Totale
 Nrj=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
     {
      us_l=Img[0][i][j];
      us_a=Img[1][i][j];
      us_b=Img[2][i][j];
      //ValLabRgbLab(&us_l,&us_a,&us_b);

      usl=(int)(us_l/FACT_CUBE);
      usa=(int)(us_a/FACT_CUBE);
      usb=(int)(us_b/FACT_CUBE);
      if (usl>=LMAX) usl=LMAX-1;
      if (usl<LMIN)  usl=LMIN;
      if (usa>=AMAX) usa=AMAX-1;
      if (usa<AMIN)  usa=AMIN;
      if (usb>=BMAX) usb=BMAX-1;
      if (usb<BMIN)  usb=BMIN;
      us_l=CubLabRgbLab[usl-LMIN][usa-AMIN][usb-BMIN][0];
      us_a=CubLabRgbLab[usl-LMIN][usa-AMIN][usb-BMIN][1];
      us_b=CubLabRgbLab[usl-LMIN][usa-AMIN][usb-BMIN][2];

      tmp0=0.0;
      tmp1=0.0;
      tmp2=0.0;
      for(k=0;k<NbCnx;k++)
        { nbp=(i*wdth)+j;
          posr=(int)NonLocGrph[0][k][nbp];
          posc=(int)NonLocGrph[1][k][nbp];
          dist=NonLocGrph[2][k][nbp];

          ut_l=Img[0][posr][posc];
          ut_a=Img[1][posr][posc];
          ut_b=Img[2][posr][posc];
          //ValLabRgbLab(&ut_l,&ut_a,&ut_b);

          utl=(int)(ut_l/FACT_CUBE);
          uta=(int)(ut_a/FACT_CUBE);
          utb=(int)(ut_b/FACT_CUBE);
          if (utl>=LMAX) utl=LMAX-1;
          if (utl<LMIN)  utl=LMIN;
          if (uta>=AMAX) uta=AMAX-1;
          if (uta<AMIN)  uta=AMIN;
          if (utb>=BMAX) utb=BMAX-1;
          if (utb<BMIN)  utb=BMIN;
          ut_l=CubLabRgbLab[utl-LMIN][uta-AMIN][utb-BMIN][0];
          ut_a=CubLabRgbLab[utl-LMIN][uta-AMIN][utb-BMIN][1];
          ut_b=CubLabRgbLab[utl-LMIN][uta-AMIN][utb-BMIN][2];
    
          tmp0=CARRE(us_l-ut_l)+CARRE(us_a-ut_a)+CARRE(us_b-ut_b);
          tmp1=CARRE(fact*dist-sqrt(tmp0));
          tmp2+=tmp1-(Regul*tmp0); }

      Nrj+=tmp2/(float)NbCnx;
     }

 //Retour
 return Nrj/(float)(wdth*lgth);
}
